// Octree.cpp: implementation of the COctree class.
//
//	C++ class written by Brett Porter
//	Based on code by Dean Clark for Dr. Dobbs Journal, May 1995
//////////////////////////////////////////////////////////////////////

#include "stdpch.h"
#include "bitmap.h"
#include "Octree.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

COctree::COctree( DWORD nColours )
{
	m_LeafLevel = TREEDEPTH;
	m_TotalLeaves = 0;
	m_nMaxColours = nColours;

	for ( int i = 0; i < TREEDEPTH; i++ )
		m_pReduceList[i] = NULL;

	m_pRoot = NULL;
}

COctree::~COctree()
{
	if ( m_pRoot != NULL )
		delete m_pRoot;
}

void	COctree::InsertColour( RGBAPIXEL colour )
{
	InsertNodeAt( m_pRoot, colour, 0 );

	while ( m_TotalLeaves > m_nMaxColours )
		ReduceTree();
}

COctree::COctNode	*COctree::CreateOctNode( int level )
{
	BOOL		leaf = ( level == m_LeafLevel ) ? TRUE : FALSE;
    COctNode	*newNode = new COctNode( level, leaf );

	if ( leaf )
		m_TotalLeaves++;
	else
		MakeReducible( level, newNode );

	return newNode;
}

COctree::COctNode	*COctree::GetReducible()
{
	COctNode	*pNode;

	while ( m_pReduceList[m_LeafLevel-1] == NULL )
		m_LeafLevel--;

	pNode = m_pReduceList[m_LeafLevel-1];
	m_pReduceList[m_LeafLevel-1] = m_pReduceList[m_LeafLevel-1]->GetNextNode();

	return pNode;
}

//////////////////////////////////////////////////////////////////////
// COctNode class
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

COctree::COctNode::COctNode( int level, BOOL leaf )
{
	m_Level = level;
	m_IsLeaf = leaf;
	m_nPixels = 0;
	m_Index = 0;
	m_RedSum = m_GreenSum = m_BlueSum = 0;
	for ( int i = 0; i < COLOURBITS; i++ )
		m_pChild[i] = NULL;
}

COctree::COctNode::~COctNode()
{
	for ( int i = 0; i < COLOURBITS; i++ )
		if ( m_pChild[i] != NULL )
			delete m_pChild[i];
}

void	COctree::InsertNodeAt( COctNode *&pNode, RGBAPIXEL colour, UINT depth )
{
	if ( pNode == NULL )
		pNode = CreateOctNode( depth );

	if ( pNode->IsLeaf())
		pNode->AddPixel(( BYTE* )&colour );
	else
	{
		int	level = COctNode::GetLevel(( BYTE* )&colour, TREEDEPTH-depth );
		InsertNodeAt( pNode->GetChild( level ), colour, depth+1 );
	}
}

int	COctree::COctNode::Reduce()
{
	DWORD		sumred = 0, sumgreen = 0, sumblue = 0;
	int			i, nChild = 0;

	for ( i = 0; i < COLOURBITS; i++ )
	{
		if ( m_pChild[i] != NULL )
		{
			nChild++;
			sumred += m_pChild[i]->m_RedSum;
			sumgreen += m_pChild[i]->m_GreenSum;
			sumblue += m_pChild[i]->m_BlueSum;
			m_nPixels += m_pChild[i]->m_nPixels;
			delete m_pChild[i];
			m_pChild[i] = NULL;
		}
	}
	m_IsLeaf = TRUE;
	m_RedSum = sumred;
	m_GreenSum = sumgreen;
	m_BlueSum = sumblue;

	return nChild;
}

void	COctree::COctNode::MakePaletteTable( RGBAPIXEL pTable[], int &index )
{
	int	i;

	if ( m_IsLeaf )
	{
		BYTE	*pColour = ( BYTE* )&pTable[index];
		pColour[RGBA_RED] = ( BYTE )( m_RedSum/m_nPixels );
		pColour[RGBA_GREEN] = ( BYTE )( m_GreenSum/m_nPixels );
		pColour[RGBA_BLUE] = ( BYTE )( m_BlueSum/m_nPixels );
		pColour[RGBA_ALPHA] = 255;	// Opaque
		m_Index = index;
		index++;
	}
	else
		for ( i = 0; i < COLOURBITS; i++ )
			if ( m_pChild[i] != NULL )
				m_pChild[i]->MakePaletteTable( pTable, index );
}

